﻿/*
	Paintable Computing Simulation

	License

	This software (Paintable Computing Simulation and all parts of its 
	source code) is provided "as is", without any warranty and it is 
	the property of its respective authors. Use of this software is 
	free to everyone, except in commercial and/or military purposes 
	without written permission of its authors. Disassembling the 
	executable code of this software is forbidden for any purpose 
	without written permission of its authors.

	If you intend to use any part of the source code of this software 
	or to develop this software any further, you must include this 
	license in every source file and every About box, without changing 
	it; only appending the license is allowed. And if you use any part 
	of the source code of this software, or develop it further, please 
	send an e-mail to its authors.

	Authors of this software are not responsible for any consequences 
	which use of this software, use of any part of its source or its 
	further development may cause.

	All software and parts of source code used in the development of 
	this software are property of their respective owners.

	Good luck,
	Authors

	Authors (alphabetically):
	Fabijan Lukin
		fabijan@gmail.com
		fabijan.lukin@fer.hr
	Fran Pregernik
		fran.pregernik@gmail.com
		fran.pregernik@fer.hr
	Tomislav Sukser
		tomislav.sukser@gmail.com
		tomislav.sukser@zg.t-com.hr
		tomislav.sukser@fer.hr


	Note:
	Paintable computing was firstly introduced in Programming a 
	Paintable Computer written by William Joseph Butera.


    Copyright (c) 2005,2006. All rights reserved.
    4-6/2006

 */
using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using PaintableSimulator.Engine.PaintableComputers.OperatingSystem;


namespace PaintableSimulator.Engine.PaintableComputers
{

	public partial class GenericPaintableComputer
	{

		#region HomePage: Write


		/// <summary>
		/// Post a string to the HomePage of the specified process fragment at the end.
		/// </summary>
		/// <param name="processFragmentId">ID of the pfrag to post to.</param>
		/// <param name="post">A string containing data.</param>
		/// <returns>
		/// Index of the post or -1 if posting failed.
		/// </returns>
		public override int PostEntry(uint processFragmentId, object post)
		{

#if TRACE
			String data = "{0:X}-{1:X}: POSTENTRY[{2}]={3}";
			if (log.IsDebugEnabled)
				log.Debug(String.Format(data, this.computerId, currentProcessFragment.ProcessFragment.ProcessFragmentId, processFragmentId, post));
#endif

			HomepagePostCollection _data;

			_data = localHomepage[processFragmentId];

			// add the post at the end of the list
			_data.Add(post);
			int index = _data.Count - 1;

			localHomepage.HasChanged = true;

			// return the index of the post
			return index;

		}


		/// <summary>
		/// Post a string to the HomePage of the current process fragment at the end.
		/// </summary>
		/// <param name="post">A string containing data.</param>
		/// <returns>
		/// Index of the post or -1 if posting failed.
		/// </returns>
		public override int PostEntry(object post)
		{
			return PostEntry(currentProcessFragment.ProcessFragment.ProcessFragmentId, post);
		}

		/// <summary>
		/// Posts a string to the HomePage of the specified process fragment at the specified index.
		/// </summary>
		/// <param name="processFragmentId">ID of the posting pfrag.</param>
		/// <param name="index">Index of the post to replace</param>
		/// <param name="post">A string containing data.</param>
		/// <returns>
		/// Index of the post or -1 if posting failed.
		/// </returns>
		public override int PostEntry(uint processFragmentId, int index, object post)
		{

#if TRACE
			String stdata = "{0:X}-{1:X}: POSTENTRY[{2},{3}]={4}";
			if (log.IsDebugEnabled)
				log.Debug(String.Format(stdata, this.computerId, currentProcessFragment.ProcessFragment.ProcessFragmentId, processFragmentId, index, post));
#endif

			HomepagePostCollection data;

			data = localHomepage[processFragmentId];


			// check bounds
			if (index < 0)
				return 0;

			// fill the difference with null's

			if (index >= data.Count)
			{
				// limit - removes sidefect for evaluation in for-loop
				int limit = index + 1 - data.Count;
				for (int i = 0; i < limit; i++)
					data.Add(null);
			}

			// change the post at the index
			data[index] = post;

			// return the index of the post
			return index;

		}

		/// <summary>
		/// Posts a string to the HomePage of the current process fragment at the specified index.
		/// </summary>
		/// <param name="index">Index of the post to replace</param>
		/// <param name="post">A string containing data.</param>
		/// <returns>
		/// Index of the post or -1 if posting failed.
		/// </returns>
		public override int PostEntry(int index, object post)
		{
			return PostEntry(currentProcessFragment.ProcessFragment.ProcessFragmentId, index, post);
		}


		/// <summary>
		/// Removes a post from the HomePage of the specified process fragment.
		/// </summary>
		/// <param name="processFragmentId">Id of the posting pfrag.</param>
		/// <param name="index">Index of the post</param>
		/// <returns>True if post removal succeded.</returns>
		public override bool RemoveEntry(uint processFragmentId, int index)
		{

#if TRACE
			String stdata = "{0:X}-{1:X}: REMOVEENTRY({2}, {3})";
			if (log.IsDebugEnabled)
				log.Debug(String.Format(stdata, this.computerId, currentProcessFragment.ProcessFragment.ProcessFragmentId, processFragmentId, index));
#endif

			if ((localHomepage[processFragmentId].Count < index) || (index < 0))
			{
				localHomepage[processFragmentId].RemoveAt(index);
				return true;
			}
			
			return false;
		}

		/// <summary>
		/// Removes a post from the HomePage of the current process fragment.
		/// </summary>
		/// <param name="index">Index of the post</param>
		/// <returns>True if the post removal succeded.</returns>
		public override bool RemoveEntry(int index)
		{
			return RemoveEntry(currentProcessFragment.ProcessFragment.ProcessFragmentId, index);
		}


		/// <summary>
		/// Removes all posts of the specified process fragment.
		/// </summary>
		/// <param name="processFragmentId">ID of the Process Fragment</param>
		/// <returns>The number of posts removed.</returns>
		public override int RemoveAllEntries(uint processFragmentId)
		{

#if TRACE
			String data = "{0:X}-{1:X}: REMOVEALLENTRIES({2})";
			if (log.IsDebugEnabled)
				log.Debug(String.Format(data, this.computerId, currentProcessFragment.ProcessFragment.ProcessFragmentId, processFragmentId));
#endif

			int count = 0;
			count = localHomepage[processFragmentId].Count;
			localHomepage[processFragmentId].Clear();

			for (int i=0; i<installedProcessFragments.Count; i++)
				if ((installedProcessFragments[i] != currentProcessFragment) && (installedProcessFragments[i].ProcessFragmentType == currentProcessFragment.ProcessFragmentType))
					return count;


			// if he is the last of his kind delete his personal homepage
			localHomepage.Remove(processFragmentId);
			return count;

		}

		/// <summary>
		/// Removes all posts of the current process fragment.
		/// </summary>
		/// <returns>The number of posts removed.</returns>
		public override int RemoveAllEntries()
		{
			return RemoveAllEntries(currentProcessFragment.ProcessFragment.ProcessFragmentId);
		}

		#endregion

		#region HomePage: Read


		/// <summary>
		/// Returns the number of posts of the specified process fragment.
		/// </summary>
		/// <param name="processFragmentId">ID of the Process Fragment</param>
		/// <returns></returns>
		public override int GetPostCount(uint processFragmentId)
		{

#if TRACE
			String data = "{0:X}-{1:X}: GETPOSTCOUNT({2}) == {3}";
			if (log.IsDebugEnabled)
				log.Debug(String.Format(data, this.computerId, currentProcessFragment.ProcessFragment.ProcessFragmentId, processFragmentId, localHomepage[processFragmentId].Count));
#endif
			if (localHomepage.ContainsKey(processFragmentId))
				return localHomepage[processFragmentId].Count;
			else
				return 0;
		}

		/// <summary>
		/// Returns the number of posts of the current process fragment.
		/// </summary>
		/// <returns></returns>
		public override int GetPostCount()
		{
			return GetPostCount(currentProcessFragment.ProcessFragment.ProcessFragmentId);
		}


		/// <summary>
		/// Returns the post value of the specified process fragment.
		/// </summary>
		/// <param name="processFragmentId">ID of the Process Fragment</param>
		/// <param name="index">Index of the post</param>
		/// <returns>
		/// A Object containing data if it exists; null otherwise
		/// </returns>
		public override object GetPost(uint processFragmentId, int index)
		{
			Object retval = null;

			if ((localHomepage[processFragmentId].Count > index) && (index >= 0))
				retval = localHomepage[processFragmentId][index];

#if TRACE
			String data = "{0:X}-{1:X}: GETPOST({2}, {3}) == {4}";
			if (log.IsDebugEnabled)
				log.Debug(String.Format(data, this.computerId, currentProcessFragment.ProcessFragment.ProcessFragmentId, processFragmentId, index, retval));
#endif


			return retval;
		}


		/// <summary>
		/// Returns the post value of the current process fragment.
		/// </summary>
		/// <param name="index">Index of the post</param>
		/// <returns>
		/// A object containing data if it exists; null otherwise
		/// </returns>
		public override object GetPost(int index)
		{
			return GetPost(currentProcessFragment.ProcessFragment.ProcessFragmentId, index);
		}

		#endregion

		#region I/O Space: Read


		/// <summary>
		/// This method returns the number of neighboring homepages mirrored in the IO space.
		/// </summary>
		/// <returns>The number of mirrored homepages.</returns>
		public override int GetIOSpaceCount()
		{

#if TRACE
			String data = "{0:X}-{1:X}: GETIOSPACECOUNT == {2}";
			if (log.IsDebugEnabled)
				log.Debug(String.Format(data, this.computerId, currentProcessFragment.ProcessFragment.ProcessFragmentId, mirroredHomepages.Count));
#endif

			return mirroredHomepages.Count;
		}


		/// <summary>
		/// Returns the number of posts in the chosen process fragment mirror space.
		/// </summary>
		/// <param name="mirrorIndex">Index of the mirror</param>
		/// <param name="processFragmentId">ID of the Process Fragment</param>
		/// <returns>
		/// A count of posts or 0 if the specified process fragment has no posts, -1 if the mirror doesn't exist
		/// </returns>
		public override int GetIOPostCount(int mirrorIndex, uint processFragmentId)
		{
			int count = -1;

			// ToDo: very slow
			if ((mirrorIndex >= 0) && (mirrorIndex < mirroredHomepages.Count))
			{
				Homepage hp = mirroredHomepages[mirrorIndex].Homepage;
				if (hp.ContainsKey(processFragmentId))
					count = hp[processFragmentId].Count;
				else
					count = 0;
			}
			else
				count = -1;

#if TRACE
			String data = "{0:X}-{1:X}: GETIOPOSTCOUNT({2}, {3}) == {4}";
			if (log.IsDebugEnabled)
				log.Debug(String.Format(data, this.computerId, currentProcessFragment.ProcessFragment.ProcessFragmentId, mirrorIndex, processFragmentId, count));
#endif

			return count;

		}

		/// <summary>
		/// Returns the number of posts in the mirror space for the current process fragment type.
		/// </summary>
		/// <param name="mirrorIndex">Index of the mirror</param>
		/// <returns>
		/// A count of posts or 0 if the specified process fragment has no posts, -1 if the mirror doesn't exist
		/// </returns>
		public override int GetIOPostCount(int mirrorIndex)
		{
			return GetIOPostCount(mirrorIndex, currentProcessFragment.ProcessFragment.ProcessFragmentId);
		}


		/// <summary>
		/// Returns a mirrored post value of the specified process fragment.
		/// </summary>
		/// <param name="mirrorIndex">Index of the mirrored HomePage.</param>
		/// <param name="processFragmentId">Guid of the PFrag</param>
		/// <param name="postIndex">Index of the post.</param>
		/// <returns>The post if it exists or null otherwise</returns>
		public override object GetIOPost(int mirrorIndex, uint processFragmentId, int postIndex)
		{

			Object retval = null;

			if (mirroredHomepages[mirrorIndex] != null)
			{
				ExtendedHomepageData ehp = mirroredHomepages[mirrorIndex];
				HomepagePostCollection posts = ehp.Homepage[processFragmentId];

				if (posts != null)
				{
					if ((postIndex >= 0) && (postIndex < posts.Count))
						retval = posts[postIndex];
					else
						retval = null;
				}
				else
					retval = null;
			}
			else
				retval = null;


#if TRACE
			String data = "{0:X}-{1:X}: GETIOPOST({2}, {3}, {4}) == {5}";
			if (log.IsDebugEnabled)
				log.Debug(String.Format(data, this.computerId, currentProcessFragment.ProcessFragment.ProcessFragmentId, mirrorIndex, processFragmentId, postIndex, retval));
#endif

			return retval;

		}

		/// <summary>
		/// Returns a mirrored post value of the current process fragment type.
		/// </summary>
		/// <param name="mirrorIndex">Index of the mirror.</param>
		/// <param name="postIndex">Index of the post.</param>
		/// <returns></returns>
		public override object GetIOPost(int mirrorIndex, int postIndex)
		{
			return GetIOPost(mirrorIndex, currentProcessFragment.ProcessFragment.ProcessFragmentId, postIndex);
		}

		#endregion

		#region Process Fragment Transfer


		/// <summary>
		/// Queue a process fragment transfer ticket.
		/// </summary>
		/// <param name="mirrorIndex">The index of the mirrored HomePage of the paintable computer to transfer to.</param>
		/// <returns>True if queuing succeded</returns>
		public override bool QueueTransfer(int mirrorIndex)
		{
			return QueueTransfer(mirrorIndex, 0);
		}


		/// <summary>
		/// Queue a prioritized process fragment transfer ticket.
		/// </summary>
		/// <param name="mirrorIndex">The index of the mirrored Homepage of the Paintable to transfer to.</param>
		/// <param name="priority">Priority</param>
		/// <returns></returns>
		public override bool QueueTransfer(int mirrorIndex, byte priority)
		{
			if (currentProcessFragment.QueuedForUninstall)
				return false;

			if (currentProcessFragment.IsQueuedForTransfer(mirrorIndex))
				return false;

#if TRACE
			String data = "{0:X}-{1:X}: QUEUETRANSFER({2}, {3})";
			if (log.IsDebugEnabled)
				log.Debug(String.Format(data, this.computerId, currentProcessFragment.ProcessFragment.ProcessFragmentId, mirrorIndex, priority));
#endif

			// queue request for further processing
			preTransferQueue.Enqueue(new TransferRequestTicket(priority, mirrorIndex, mirroredHomepages[mirrorIndex].OriginComputerId, currentProcessFragment));
			currentProcessFragment.SetQueuedForTransferFlag(mirrorIndex);

			return true;
		}



		/// <summary>
		/// Used to check if any transfer requests are still pending.
		/// </summary>
		/// <returns>True if requests are still pending</returns>
		public override bool IsQueuedForTransfer()
		{

#if TRACE
			String data = "{0:X}-{1:X}: ISQUEUEDFORTRANSFER";
			if (log.IsDebugEnabled)
				log.Debug(String.Format(data, this.computerId, currentProcessFragment.ProcessFragment.ProcessFragmentId));
#endif

			if (sendPadQueue.Count == 0)
				return false;


			if (sendPadQueue.Peek().PacketType != PacketType.ProcessFragment)
				return false;

			// loop through table, find the pfrag and check if queued
			for (int i = 0; i < maximumNeighbourCount; i++)
				if (currentProcessFragment.IsQueuedForTransfer(i))
					return true;

			// no requests queued
			return false;

		}



		/// <summary>
		/// Used to check if a specific transfer request to the selected mirror is still pending.
		/// </summary>
		/// <param name="mirrorIndex"></param>
		/// <returns>True if request is still pending</returns>
		public override bool IsQueuedForTransfer(int mirrorIndex)
		{

#if TRACE
			String data = "{0:X}-{1:X}: ISQUEUEDFORTRANSFER({2}";
			if (log.IsDebugEnabled)
				log.Debug(String.Format(data, this.computerId, currentProcessFragment.ProcessFragment.ProcessFragmentId, mirrorIndex));
#endif

			if (sendPadQueue.Count == 0)
				return false;


			if (sendPadQueue.Peek().PacketType != PacketType.ProcessFragment)
				return false;

			// loop through table, find the pfrag and check if queued

			return currentProcessFragment.IsQueuedForTransfer(mirrorIndex);
		}


		/// <summary>
		/// Removes the pending transfer ticket to the specified mirror.
		/// </summary>
		/// <param name="mirrorIndex">The index of the mirrored Homepage</param>
		/// <returns>
		/// True if the request ticket has been removed. False if no request exists.
		/// </returns>
		public override bool ClearQueueTransfer(int mirrorIndex)
		{

#if TRACE
			String data = "{0:X}-{1:X}: CLEARQUEUETRANSFER({2})";
			if (log.IsDebugEnabled)
				log.Debug(String.Format(data, this.computerId, currentProcessFragment.ProcessFragment.ProcessFragmentId, mirrorIndex));
#endif

			// ne vidim svrhu ovoga, mislim to može OS automatski, èim se dozvoli prijenos!
			return true;
		}


		/// <summary>
		/// Schedules the current executing process fragment for uninstallation.
		/// </summary>
		/// <returns></returns>
		public override bool QueueUninstall()
		{
			return QueueUninstall(0);
		}



		/// <summary>
		/// Schedules the current executing PFrag for uninstall.
		/// </summary>
		/// <param name="priority">The priority.</param>
		/// <returns>
		/// True if allowed to transfer; false othervise;
		/// </returns>
		public override bool QueueUninstall(byte priority)
		{

#if TRACE
			String data = "{0:X}-{1:X}: QUEUEUNINSTALL({2})";
			if (log.IsDebugEnabled)
				log.Debug(String.Format(data, this.computerId, currentProcessFragment.ProcessFragment.ProcessFragmentId, priority));
#endif

			if (currentProcessFragment.QueuedForUninstall)
				return false;
			else
			{
				uninstallQueue.Push(priority, currentProcessFragment);
				currentProcessFragment.QueuedForUninstall = true;
			}

			return true;
		}


		#endregion

		#region Maintenance & Miscelaneous

		/// <summary>
		/// Gets the paintable computer Id.
		/// </summary>
		/// <returns>The Computer Id</returns>
		/// <remarks>Not guaranteed to be unique for all computers.</remarks>
		public override uint GetComputerId()
		{
			return this.ComputerId;
		}


		#endregion

	}
}
