/*
	Paintable Computing Simulation

	License

	This software (Paintable Computing Simulation and all parts of its 
	source code) is provided "as is", without any warranty and it is 
	the property of its respective authors. Use of this software is 
	free to everyone, except in commercial and/or military purposes 
	without written permission of its authors. Disassembling the 
	executable code of this software is forbidden for any purpose 
	without written permission of its authors.

	If you intend to use any part of the source code of this software 
	or to develop this software any further, you must include this 
	license in every source file and every About box, without changing 
	it; only appending the license is allowed. And if you use any part 
	of the source code of this software, or develop it further, please 
	send an e-mail to its authors.

	Authors of this software are not responsible for any consequences 
	which use of this software, use of any part of its source or its 
	further development may cause.

	All software and parts of source code used in the development of 
	this software are property of their respective owners.

	Good luck,
	Authors

	Authors (alphabetically):
	Fabijan Lukin
		fabijan@gmail.com
		fabijan.lukin@fer.hr
	Fran Pregernik
		fran.pregernik@gmail.com
		fran.pregernik@fer.hr
	Tomislav Sukser
		tomislav.sukser@gmail.com
		tomislav.sukser@zg.t-com.hr
		tomislav.sukser@fer.hr


	Note:
	Paintable computing was firstly introduced in Programming a 
	Paintable Computer written by William Joseph Butera.


    Copyright (c) 2005,2006. All rights reserved.
    4-6/2006

 */
using System;
using System.Collections.Generic;
using System.Text;
using PaintableSimulator.Engine.PaintableComputers.Sensors;
using PaintableSimulator.Engine.PaintableComputers.OperatingSystem;
using System.Drawing;

namespace PaintableSimulator.Engine.PaintableComputers.Sensors
{


	#region Colorizer

	/// <summary>
	/// The homepage colorizer of the temperature sensor
	/// </summary>
	public class TemperatureSensorHomepageColorizer : HomepageColorizer
	{
		private static double maximumTemperature = 1.0;

		/// <summary>
		/// Colorizes the specified homepage.
		/// </summary>
		/// <param name="homepage">The homepage.</param>
		/// <returns></returns>
		public override Color Colorize(Homepage homepage)
		{
			Color retval = Color.WhiteSmoke;

			for (int i = 0; i < homepage.Count; i++)
				if (homepage.KeyList[i] == TemperatureSensorPaintableComputer.TemperatureSensorProcessFragmentId)
				{
					if (homepage.DataList[i].Count == 0)
						break;

					// post at index 0 is the temperature
					double temperatureValue = (double)homepage.DataList[i][0];

					if (maximumTemperature < temperatureValue)
						maximumTemperature = temperatureValue;


					int color = (int)(255 * temperatureValue / maximumTemperature);

					retval = Color.FromArgb(color, 0, 0);

					break;
				}

			return retval;
		}
	}

	#endregion



	/// <summary>
	/// This is a PaintableComputer with a temperature sensor on board
	/// </summary>
	[DefaultHomepageColorizer(typeof(TemperatureSensorHomepageColorizer))]
	public class TemperatureSensorPaintableComputer : GenericPaintableComputer, ISensor
	{

		private double temperature;
		public const uint TemperatureSensorProcessFragmentId = 0x11111110;
				

		/// <summary>
		/// This sensor only works with the simpleTemperatureDataFeeder
		/// </summary>
		private readonly Type[] usesDataFeeders = new Type[] {typeof(SimpleTemperatureDataFeeder)};

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the <see cref="T:TemperatureSensorPaintableComputer"/> class.
		/// </summary>
		public TemperatureSensorPaintableComputer()
			: base(DEFAULT_MAXIMUMNEIGHBOURCOUNT)
		{
			// install the temperature sensors private homepage on this PaC
			localHomepage.Add(TemperatureSensorProcessFragmentId, new HomepagePostCollection());
		}


		/// <summary>
		/// Initializes a new instance of the <see cref="T:TemperatureSensorPaintableComputer"/> class.
		/// </summary>
		/// <param name="maximumNeighbourCount">The maximum neighbour count.</param>
		public TemperatureSensorPaintableComputer(uint maximumNeighbourCount)
			: base(maximumNeighbourCount)
		{
			// install the lightItensity sensors private homepage on tthis PaC
			localHomepage.Add(TemperatureSensorProcessFragmentId, new HomepagePostCollection());
		}

		#endregion
		
		#region ISensor Members

		/// <summary>
		/// Sets the environment value.
		/// </summary>
		/// <param name="dataFeederType">Type of the data feeder.</param>
		/// <param name="sensorValue">The sensor value.</param>
		public void SetEnvironmentValue(Type dataFeederType, double sensorValue)
		{
			if (dataFeederType == typeof(SimpleTemperatureDataFeeder))
			{
				temperature = sensorValue;
			}
		}

		/// <summary>
		/// Gets the data feeders this sensor works with.
		/// </summary>
		/// <value>The data feeders.</value>
		public Type[] DataFeeders
		{
			get
			{
				return usesDataFeeders;
			}
		}

		#endregion


		/// <summary>
		/// Cycles this instance.
		/// </summary>
		public override void Cycle()
		{

			// update the homepage with fresh sensor values
			UpdateSensorHomepage();

			// 
			base.Cycle();

		}



		/// <summary>
		/// Updates the sensor homepage.
		/// </summary>
		private void UpdateSensorHomepage()
		{
			// assign a arbitrary PFragId to the each sensor type:
			// e.g. TemperatureSensor is a "pfrag" with ProcessFragmentId=0x11111110
			// and write to the homepage under that ProcessFragmentId
			
			base.PostEntry(TemperatureSensorProcessFragmentId, 0, temperature);
		}


	}
}
