/*
	Paintable Computing Simulation

	License

	This software (Paintable Computing Simulation and all parts of its 
	source code) is provided "as is", without any warranty and it is 
	the property of its respective authors. Use of this software is 
	free to everyone, except in commercial and/or military purposes 
	without written permission of its authors. Disassembling the 
	executable code of this software is forbidden for any purpose 
	without written permission of its authors.

	If you intend to use any part of the source code of this software 
	or to develop this software any further, you must include this 
	license in every source file and every About box, without changing 
	it; only appending the license is allowed. And if you use any part 
	of the source code of this software, or develop it further, please 
	send an e-mail to its authors.

	Authors of this software are not responsible for any consequences 
	which use of this software, use of any part of its source or its 
	further development may cause.

	All software and parts of source code used in the development of 
	this software are property of their respective owners.

	Good luck,
	Authors

	Authors (alphabetically):
	Fabijan Lukin
		fabijan@gmail.com
		fabijan.lukin@fer.hr
	Fran Pregernik
		fran.pregernik@gmail.com
		fran.pregernik@fer.hr
	Tomislav Sukser
		tomislav.sukser@gmail.com
		tomislav.sukser@zg.t-com.hr
		tomislav.sukser@fer.hr


	Note:
	Paintable computing was firstly introduced in Programming a 
	Paintable Computer written by William Joseph Butera.


    Copyright (c) 2005,2006. All rights reserved.
    4-6/2006

 */
using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Reflection;
using PaintableSimulator.Engine.PaintableComputers.OperatingSystem;
using PaintableSimulator.Engine.PaintableComputers;
using PaintableSimulator.Engine.PaintableComputers.Sensors;
using System.Xml;

namespace PaintableSimulator.Engine
{
	/// <summary>
	/// 
	/// </summary>
	public class PluginLoader : MarshalByRefObject
	{
		private const string PluginsDirectory = "\\Plugins";

		#region Private fields

		/// <summary>
		/// Holds all valid available process fragments
		/// </summary>
		private List<FriendlyType> validProcessFragments;

		/// <summary>
		/// Holds all valid environmental data feeders
		/// </summary>
		private List<FriendlyType> validEnvironmentDataFeeders;

		/// <summary>
		/// Holds all valid sensors
		/// </summary>
		private List<FriendlyType> validSensors;

		/// <summary>
		/// Holds all valid found paintable computers
		/// </summary>
		private List<FriendlyType> validPaintableComputers;

		/// <summary>
		/// Holds all valid found homepage colorizers
		/// </summary>
		private List<FriendlyType> validHomepageColorizers;

		/// <summary>
		/// Holds a dictionary of processfragment Id to string translations
		/// </summary>
		private Dictionary<uint, string> processFragmentIdTranslations;

		/// <summary>
		/// 
		/// </summary>
		private static Dictionary<string, List<string>> blockedFunctionCalls = new Dictionary<string, List<string>>();

		/// <summary>
		/// 
		/// </summary>
		private static DirectoryInfo pluginsDirectory;

		#endregion

		#region Public Properties

		/// <summary>
		/// Gets the valid homepage colorizers.
		/// </summary>
		/// <value>The valid homepage colorizers.</value>
		public List<FriendlyType> ValidHomepageColorizers
		{
			get { return validHomepageColorizers; }
		}

		/// <summary>
		/// Gets the process fragment id translations.
		/// </summary>
		/// <value>The process fragment id translations.</value>
		public Dictionary<uint, string> ProcessFragmentIdTranslations
		{
			get { return processFragmentIdTranslations; }
		}

		/// <summary>
		/// Gets the valid paintable computers.
		/// </summary>
		/// <value>The valid paintable computers.</value>
		public List<FriendlyType> ValidPaintableComputers
		{
			get { return validPaintableComputers; }
		}

		/// <summary>
		/// Gets the valid process fragments.
		/// </summary>
		/// <value>The valid process fragments.</value>
		public List<FriendlyType> ValidProcessFragments
		{
			get { return validProcessFragments; }
		}

		/// <summary>
		/// Gets the valid environment data feeders.
		/// </summary>
		/// <value>The valid environment data feeders.</value>
		public List<FriendlyType> ValidEnvironmentDataFeeders
		{
			get { return validEnvironmentDataFeeders; }
		}

		/// <summary>
		/// Gets the valid sensors.
		/// </summary>
		/// <value>The valid sensors.</value>
		public List<FriendlyType> ValidSensors
		{
			get { return validSensors; }
		}

		#endregion

		#region Constructors

		/// <summary>
		/// Initializes the <see cref="T:PluginLoader"/> type.
		/// </summary>
		static PluginLoader()
		{
			// What user is not allowed to use
			blockedFunctionCalls.Add("System.IO.Directory", new List<string>());
			blockedFunctionCalls["System.IO.Directory"].Add("*");

			// check the plugins folder
			FileInfo assemblyFileInfo = new FileInfo(Assembly.GetExecutingAssembly().Location);
			DirectoryInfo assemblyDirectory = assemblyFileInfo.Directory;

			pluginsDirectory = new DirectoryInfo(assemblyDirectory.FullName + PluginsDirectory);
			if (!pluginsDirectory.Exists)
				pluginsDirectory.Create();

			// allow dynamic loading of plugin assemblies
			AppDomain.CurrentDomain.AssemblyResolve += new ResolveEventHandler(CurrentDomain_AssemblyResolve);
			AppDomain.CurrentDomain.AssemblyLoad += new AssemblyLoadEventHandler(CurrentDomain_AssemblyLoad);

		}

		/// <summary>
		/// Initializes the <see cref="T:PluginLoader"/> class.
		/// </summary>
		public PluginLoader()
		{

			validProcessFragments = new List<FriendlyType>();
			validEnvironmentDataFeeders = new List<FriendlyType>();
			validHomepageColorizers = new List<FriendlyType>();
			validSensors = new List<FriendlyType>();
			validPaintableComputers = new List<FriendlyType>();
			processFragmentIdTranslations = new Dictionary<uint, string>();

			// scan for plugins
			ScanPlugins(pluginsDirectory);

		}

		/// <summary>
		/// Handles the AssemblyLoad event of the CurrentDomain control.
		/// </summary>
		/// <param name="sender">The source of the event.</param>
		/// <param name="args">The <see cref="T:System.AssemblyLoadEventArgs"/> instance containing the event data.</param>
		private static void CurrentDomain_AssemblyLoad(object sender, AssemblyLoadEventArgs args)
		{
			// log assembly loaded
			if (EngineLog.IsWarnEnabled)
				EngineLog.InfoFormat("Additional assembly loaded: {0}", args.LoadedAssembly.FullName);
		}

		/// <summary>
		/// Handles the AssemblyResolve event of the Current Domain.
		/// </summary>
		/// <param name="sender">The source of the event.</param>
		/// <param name="args">The <see cref="T:System.ResolveEventArgs"/> instance containing the event data.</param>
		/// <returns></returns>
		private static Assembly CurrentDomain_AssemblyResolve(object sender, ResolveEventArgs args)
		{

			// get the dlls in the directory
			FileInfo[] fragments = pluginsDirectory.GetFiles("*.dll");

			// log search attempt
			if (EngineLog.IsInfoEnabled)
				EngineLog.InfoFormat("Searching for assembly: {0}", args.Name);

			// load assemblies
			try
			{

				for (int i = 0; i < fragments.Length; i++)
				{
					if (Assembly.ReflectionOnlyLoadFrom(fragments[i].FullName).FullName == args.Name)
						return Assembly.LoadFile(fragments[i].FullName);
				}
			}
			catch (Exception ex)
			{
				// log loading failed
				if (EngineLog.IsErrorEnabled)
					EngineLog.Error("Assebmly loading failed!", ex);
			}

			// log assembly not found
			if (EngineLog.IsWarnEnabled)
				EngineLog.WarnFormat("Assebmly not found: {0}", args.Name);

			return null;
		}

		#endregion


		/// <summary>
		/// Rescans the plugins folder.
		/// </summary>
		public void RescanPlugins()
		{
			validProcessFragments.Clear();
			processFragmentIdTranslations.Clear();
			validEnvironmentDataFeeders.Clear();
			validHomepageColorizers.Clear();
			validSensors.Clear();
			validPaintableComputers.Clear();
			ScanPlugins(pluginsDirectory);

		}

		/// <summary>
		/// Reads the plugins subdirectory for ProcessFragments, sensors, environment data feeders
		/// </summary>
		private void ScanPlugins(DirectoryInfo pluginsDirectory)
		{

			// get the dlls in the directory
			System.IO.FileInfo[] fragments = pluginsDirectory.GetFiles("*.dll");


			//scan local assembly
			try
			{
				ScanAssemblyForProcessFragments(this.GetType().Assembly, ref validProcessFragments);
				ScanAssemblyForEnvironmentFeeders(this.GetType().Assembly, ref validEnvironmentDataFeeders);
				ScanAssemblyForHomepageColorizers(this.GetType().Assembly, ref validHomepageColorizers);
				ScanAssemblyForSensors(this.GetType().Assembly, ref validSensors);
				ScanAssemblyForPaintableComputers(this.GetType().Assembly, ref validPaintableComputers);
			}
			catch (Exception ex)
			{
				if (EngineLog.IsDebugEnabled)
					EngineLog.Debug("PluginLoader", ex);
			}

			// load assemblies
			for (int i = 0; i < fragments.Length; i++)
			{
				Assembly assembly = Assembly.LoadFile(fragments[i].FullName);

				//// check if the pfrag asswmbly has been signed
				//if (assembly.ManifestModule.GetSignerCertificate() == null)
				//    return;

				// find process fragments
				try
				{
					ScanAssemblyForProcessFragments(assembly, ref validProcessFragments);
				}
				catch
				{
				}

				// find paintable computers
				try
				{
					ScanAssemblyForPaintableComputers(assembly, ref validPaintableComputers);
				}
				catch
				{
				}

				// find sensors
				try
				{
					ScanAssemblyForSensors(assembly, ref validSensors);
				}
				catch
				{
				}

				// find environment data feeders
				try
				{
					ScanAssemblyForEnvironmentFeeders(assembly, ref validEnvironmentDataFeeders);
				}
				catch
				{
				}

				// find homepage colorizers
				try
				{
					ScanAssemblyForHomepageColorizers(assembly, ref validHomepageColorizers);
				}
				catch
				{
				}

				// create a dictionary of pfrag Id translations
				CreatePrcessFragmentIdTranslations();

			}

		}

		/// <summary>
		/// Scans the assembly for homepage colorizers.
		/// </summary>
		/// <param name="assembly">The assembly.</param>
		/// <param name="colorizers">The colorizers.</param>
		private void ScanAssemblyForHomepageColorizers(Assembly assembly, ref List<FriendlyType> colorizers)
		{
			Type[] types = null;

			// load all exported types from the assembly
			try
			{
				types = assembly.GetExportedTypes();
			}
			catch
			{
				throw new Exception(String.Format("Error loading types from the assembly {0}", assembly.Location));
			}

			// check types
			for (int i = 0; i < types.Length; i++)
			{

				if (!types[i].IsSubclassOf(typeof(HomepageColorizer)))
					continue;

				if (!types[i].IsPublic)
					continue;

				//The type should have a default constructor defined
				if (types[i].GetConstructor(Type.EmptyTypes) == null)
				{
					if (EngineLog.IsWarnEnabled)
						EngineLog.WarnFormat("Colorizer {0} has no deafult constructor defined", types[i]);
					continue;
				}


				if (EngineLog.IsInfoEnabled)
					EngineLog.InfoFormat("Homepage colorizer: {0} loaded.", types[i]);

				colorizers.Add(types[i]);


			}

			colorizers.Sort(new Comparison<FriendlyType>(TypeNameComparison));

		}


		/// <summary>
		/// Scans the assembly for pfrags.
		/// </summary>
		/// <param name="assembly">The assembly.</param>
		/// <param name="pfrags">The pfrags.</param>
		private void ScanAssemblyForProcessFragments(Assembly assembly, ref List<FriendlyType> pfrags)
		{

			Type[] types = null;

			// load all exported types from the assembly
			try
			{
				types = assembly.GetExportedTypes();
			}
			catch
			{
				throw new Exception(String.Format("Error loading types from the assembly {0}", assembly.Location));
			}

			// check types
			for (int i = 0; i < types.Length; i++)
			{
				try
				{
					if (CheckProcessFragmentValidity(types[i]))
					{
						if (EngineLog.IsInfoEnabled)
							EngineLog.InfoFormat("Process fragment: {0} loaded.", types[i]);

						pfrags.Add(types[i]);
					}
				}
				catch
				{
					continue;
				}
			}

			pfrags.Sort(new Comparison<FriendlyType>(TypeNameComparison));

		}

		/// <summary>
		/// Creates the translations.
		/// </summary>
		private void CreatePrcessFragmentIdTranslations()
		{
			for (int i = 0; i < validProcessFragments.Count; i++)
			{
				try
				{
					// create a pfrag
					ProcessFragmentBase pfrag = (ProcessFragmentBase)Activator.CreateInstance(validProcessFragments[i].InnerType);
					processFragmentIdTranslations.Add(pfrag.ProcessFragmentId, validProcessFragments[i].UserDefinedName);
				}
				catch
				{
					//just skip this one
				}
			}

			// TODO: Somehow add the sensors ID's - very difficult
		}


		/// <summary>
		/// Scans the assembly for environment feeders.
		/// </summary>
		/// <param name="assembly">The assembly.</param>
		/// <param name="feeders">The feeders.</param>
		private void ScanAssemblyForEnvironmentFeeders(Assembly assembly, ref List<FriendlyType> feeders)
		{

			Type[] types = null;

			// load all exported types from the assembly
			try
			{
				types = assembly.GetExportedTypes();
			}
			catch
			{
				throw new Exception(String.Format("Error loading types from the assembly {0}", assembly.Location));
			}

			// check types
			for (int i = 0; i < types.Length; i++)
			{

				if (!types[i].IsSubclassOf(typeof(EnvironmentDataFeeder)))
					continue;

				if (!types[i].IsPublic)
					continue;

				//The type should have a constructor defined with the parameter System.Drawing.Size
				if (types[i].GetConstructor(new Type[] { typeof(System.Drawing.Size) }) == null)
				{
					if (EngineLog.IsWarnEnabled)
						EngineLog.WarnFormat("Environment data feeder {0} has no constructor Feeder(System.Drawing.Size) defined", types[i]);
					continue;
				}

				if (EngineLog.IsInfoEnabled)
					EngineLog.InfoFormat("Environment data feeder: {0} loaded.", types[i]);


				feeders.Add(types[i]);


			}

			feeders.Sort(new Comparison<FriendlyType>(TypeNameComparison));

		}

		/// <summary>
		/// Scans the assembly for sensors.
		/// </summary>
		/// <param name="assembly">The assembly.</param>
		/// <param name="sensors">The sensors.</param>
		private void ScanAssemblyForSensors(Assembly assembly, ref List<FriendlyType> sensors)
		{

			Type[] types = null;

			// load all exported types from the assembly
			try
			{
				types = assembly.GetExportedTypes();
			}
			catch
			{
				throw new Exception(String.Format("Error loading types from the assembly {0}", assembly.Location));
			}

			// check types
			for (int i = 0; i < types.Length; i++)
			{

				if (types[i].GetInterface(typeof(ISensor).Name) == null)
					continue;

				if (!types[i].IsSubclassOf(typeof(PaintableComputerBase)))
					continue;
				
				if (!types[i].IsPublic)
					continue;

				//The type should have a constructor defined with the parameter uint - maximum neighbour count
				if (types[i].GetConstructor(new Type[] { typeof(uint) }) == null)
				{
					if (EngineLog.IsWarnEnabled)
						EngineLog.WarnFormat("Sensor {0} has no constructor Sensor(uint) defined", types[i]);
					continue;
				}


				if (EngineLog.IsInfoEnabled)
					EngineLog.InfoFormat("Sensor computer: {0} loaded.", types[i]);

				sensors.Add(types[i]);

			}

			sensors.Sort(new Comparison<FriendlyType>(TypeNameComparison));

		}

		/// <summary>
		/// Scans the assembly for paintable computers including sensors.
		/// </summary>
		/// <param name="assembly">The assembly.</param>
		/// <param name="computers">The computers.</param>
		private void ScanAssemblyForPaintableComputers(Assembly assembly, ref List<FriendlyType> computers)
		{

			Type[] types = null;

			// load all exported types from the assembly
			try
			{
				types = assembly.GetExportedTypes();
			}
			catch
			{
				throw new Exception(String.Format("Error loading types from the assembly {0}", assembly.Location));
			}

			// check types
			for (int i = 0; i < types.Length; i++)
			{

				if (!types[i].IsSubclassOf(typeof(PaintableComputerBase)))
					continue;

				if (!types[i].IsPublic)
					continue;

				//The type should have a constructor defined with the parameter uint - maximum neighbour count
				if (types[i].GetConstructor(new Type[] { typeof(uint) }) == null)
				{
					if (EngineLog.IsWarnEnabled)
						EngineLog.WarnFormat("Computer {0} has no constructor Computer(uint) defined", types[i]);
					continue;
				}


				if (EngineLog.IsInfoEnabled)
					EngineLog.InfoFormat("Paintable Computer: {0} loaded.", types[i]);

				computers.Add(types[i]);

			}

			computers.Sort(new Comparison<FriendlyType>(TypeNameComparison));

		}



		/// <summary>
		/// Returns an approximate of the size of the PFrag
		/// </summary>
		/// <param name="pfragType">Type of the pfrag.</param>
		/// <returns></returns>
		private int GetProcessFragmentSize(Type pfragType)
		{
			int size = 0;

			if (pfragType.IsValueType)
				return 0;

			ConstructorInfo[] constructors = pfragType.GetConstructors();
			MethodInfo[] methods = pfragType.GetMethods();
			PropertyInfo[] properties = pfragType.GetProperties();
			Type[] nested = pfragType.GetNestedTypes();

			//get size of MSIL for constructors
			for (int i = 0; i < constructors.Length; i++)
			{
				MethodBody methodBody = constructors[i].GetMethodBody();
				size += methodBody.GetILAsByteArray().Length;
			}

			//get the size of MSIL for methods
			for (int i = 0; i < methods.Length; i++)
			{
				MethodBody methodBody = methods[i].GetMethodBody();
				size += methodBody.GetILAsByteArray().Length;
			}

			//get the size of MSIL for methods
			for (int i = 0; i < properties.Length; i++)
			{
				//getter
				MethodInfo getter = properties[i].GetGetMethod();
				if (getter != null)
				{
					MethodBody methodBody = getter.GetMethodBody();
					size += methodBody.GetILAsByteArray().Length;
				}

				//setter
				MethodInfo setter = properties[i].GetSetMethod();
				if (setter != null)
				{
					MethodBody methodBody = setter.GetMethodBody();
					size += methodBody.GetILAsByteArray().Length;
				}
			}

			//recursivly call get size for nested types
			for (int i = 0; i < nested.Length; i++)
				size += GetProcessFragmentSize(nested[i]);

			return size;
		}

		/// <summary>
		/// Checks the PFrag validity.
		/// </summary>
		/// <param name="pfrag">The pfrag.</param>
		private bool CheckProcessFragmentValidity(Type pfrag)
		{
			//The type should inherit from ProcessFragment
			if (!(pfrag.BaseType == typeof(ProcessFragmentBase)))
				return false;

			//The type should be sealed
			if (!pfrag.IsSealed)
			{
				if (EngineLog.IsWarnEnabled)
					EngineLog.WarnFormat("Process fragment {0} is not a sealed class", pfrag);
				return false;
			}

			// it should not be abstract
			if (pfrag.IsAbstract)
			{
				if (EngineLog.IsWarnEnabled)
					EngineLog.WarnFormat("Process fragment {0} is marked as abstract", pfrag);
				return false;
			}

			// it should be serializable
			if (!pfrag.IsSerializable)
			{
				if (EngineLog.IsWarnEnabled)
					EngineLog.WarnFormat("Process fragment {0} is not serializable", pfrag);
				return false;
			}

			//The type should be public
			if (!pfrag.IsPublic)
			{
				if (EngineLog.IsWarnEnabled)
					EngineLog.WarnFormat("Process fragment {0} is not public", pfrag);
				return false;
			}

			// no implemented interfaces
			if (pfrag.GetInterfaces().Length != 0)
			{
				if (EngineLog.IsWarnEnabled)
					EngineLog.WarnFormat("Process fragment {0} implements interfaces", pfrag);
				return false;
			}

			//The type should not have any events declared
			if (pfrag.GetEvents().Length != 0)
			{
				if (EngineLog.IsWarnEnabled)
					EngineLog.WarnFormat("Process fragment {0} has events declared!", pfrag);
				return false;
			}

			//The type should have a default constructor defined
			if (pfrag.GetConstructor(Type.EmptyTypes) == null)
			{
				if (EngineLog.IsWarnEnabled)
					EngineLog.WarnFormat("Process fragment {0} has no deafult constructor defined", pfrag);
				return false;
			}

			return true;


			//////////////////////////////////////////////////////////////////////
			////                                                                //
			////     ENGINE'S MALITIUS PROGRAM EXECUTION PREVENTION (EMPEP)     //
			////                                                                //
			//////////////////////////////////////////////////////////////////////
			//ConstructorInfo[] pfragConstructors = pfrag.GetConstructors();
			//MethodInfo[] pfragMethods = pfrag.GetMethods();
			//List<MethodBody> pfragBodies = new List<MethodBody>();
			//for (int i = 0; i < pfragConstructors.Length; i++)
			//{
			//    pfragBodies.Add(pfragConstructors[i].GetMethodBody());
			//}
			//for (int i = 0; i < pfragMethods.Length; i++)
			//{
			//    pfragBodies.Add(pfragMethods[i].GetMethodBody());
			//}

			//Byte[] code;
			//string[] ilcode;
			//int[] ipcode;
			//for (int i = 0; i < pfragBodies.Count; i++)
			//{
			//    if (pfragBodies[i] == null) continue;

			//    code = pfragBodies[i].GetILAsByteArray();
			//    try
			//    {
			//        ILDisassembler.Disassemble(code, pfrag.Module, out ilcode, out ipcode);
			//    }
			//    catch
			//    {
			//        try
			//        {
			//            ILDisassembler.Disassemble(code, this.GetType().Module, out ilcode, out ipcode);
			//        }
			//        catch
			//        {
			//            continue;
			//        }
			//    }
			//    foreach (string s in ilcode)
			//    {
			//        if ((s.IndexOf("call") != -1) ||
			//            (s.IndexOf("calli") != -1) ||
			//            (s.IndexOf("callvirt") != -1))
			//        {
			//            string functionNamespace = s.Substring(s.IndexOf("] ") + 2);
			//            string functionName = functionNamespace.Substring(functionNamespace.LastIndexOf(".") + 1);
			//            functionNamespace = functionNamespace.Remove(functionNamespace.Length - functionName.Length - 1);
			//            //Sometimes, extra "." can be found on the end of namespace
			//            if (functionNamespace.Length > 0)
			//                if (functionNamespace[functionNamespace.Length - 1] == '.')
			//                    functionNamespace = functionNamespace.Remove(functionNamespace.Length - 1);

			//            if (blockedFunctionCalls.ContainsKey(functionNamespace))
			//            {
			//                List<string> namespaces = blockedFunctionCalls[functionNamespace];
			//                // Are all functions blocked?
			//                if (namespaces.IndexOf("*") != -1)
			//                    return false;
			//                // Is this function blocked?
			//                if (namespaces.IndexOf(functionName) != -1)
			//                    return false;
			//            }
			//        }
			//    }
			//}
			//////////////////////////////////////////////////////////////////////
			////                                                                //
			////     ENGINE'S MALITIUS PROGRAM EXECUTION PREVENTION (EMPEP)     //
			////                                                                //
			//////////////////////////////////////////////////////////////////////

			//return true;
		}



		/// <summary>
		/// Type name comparison.
		/// </summary>
		/// <param name="first">The first.</param>
		/// <param name="second">The second.</param>
		/// <returns></returns>
		private int TypeNameComparison(FriendlyType first, FriendlyType second)
		{
			return string.Compare(first.UserDefinedName, second.UserDefinedName);
		}

	}
}
