/*
	Paintable Computing Simulation

	License

	This software (Paintable Computing Simulation and all parts of its 
	source code) is provided "as is", without any warranty and it is 
	the property of its respective authors. Use of this software is 
	free to everyone, except in commercial and/or military purposes 
	without written permission of its authors. Disassembling the 
	executable code of this software is forbidden for any purpose 
	without written permission of its authors.

	If you intend to use any part of the source code of this software 
	or to develop this software any further, you must include this 
	license in every source file and every About box, without changing 
	it; only appending the license is allowed. And if you use any part 
	of the source code of this software, or develop it further, please 
	send an e-mail to its authors.

	Authors of this software are not responsible for any consequences 
	which use of this software, use of any part of its source or its 
	further development may cause.

	All software and parts of source code used in the development of 
	this software are property of their respective owners.

	Good luck,
	Authors

	Authors (alphabetically):
	Fabijan Lukin
		fabijan@gmail.com
		fabijan.lukin@fer.hr
	Fran Pregernik
		fran.pregernik@gmail.com
		fran.pregernik@fer.hr
	Tomislav Sukser
		tomislav.sukser@gmail.com
		tomislav.sukser@zg.t-com.hr
		tomislav.sukser@fer.hr


	Note:
	Paintable computing was firstly introduced in Programming a 
	Paintable Computer written by William Joseph Butera.


    Copyright (c) 2005,2006. All rights reserved.
    4-6/2006

 */
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.Diagnostics;
using System.Drawing;
using System.Runtime.Remoting;
using System.Threading;
using PaintableSimulator.Engine.PaintableComputers;
using PaintableSimulator.Engine.PaintableComputers.OperatingSystem;
using PaintableSimulator.Engine.Resources;
using PaintableSimulator.Engine.PaintableComputers.Sensors;
using System.Reflection;
using System.Security;

namespace PaintableSimulator.Engine
{
	public class RemoteSimulator : SimulatorBase
	{

		private IndexedDictionary<EnvironmentDataFeeder, List<int>> environment;

		#region Private fields

		private bool initialized = false;

		private uint meanTimeBetweenFailures = 0;

		/// <summary>
		/// 
		/// </summary>
		private BackgroundWorker simulationWorker;

		/// <summary>
		///
		/// </summary>
		private EngineSettings simulationSettings;

		/// <summary>
		///
		/// </summary>
		private SimulatorInfo info;


		/// <summary>
		///
		/// </summary>
		private Queue<RadioPacket> pendingRadioPackets = new Queue<RadioPacket>();

		/// <summary>
		///
		/// </summary>
		private List<SimpleComputerInfo> computerInfoList;

		/// <summary>
		///
		/// </summary>
		private List<PaintableComputerBase> computerList;

		/// <summary>
		/// 
		/// </summary>
		private long equalizationCoeficient;

		/// <summary>
		/// 
		/// </summary>
		private object syncRoot = new object();

		#endregion

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the <see cref="T:RemoteSimulator"/> class.
		/// </summary>
		public RemoteSimulator()
		{
			this.simulationWorker = new BackgroundWorker();
			//
			// simulationWorker
			//
			this.simulationWorker.WorkerSupportsCancellation = true;

			// loads the default simulation settings values
			simulationSettings = new EngineSettings();

			if (log.IsInfoEnabled)
				log.Info(UserMessages.SimulatorCreation);

			info = new SimulatorInfo();

			//
			info.simulatorState = SimulatorStates.Created;

		}

		#endregion

		#region Multithreading

		/// <summary>
		/// Handles the DoWork event of the simulationWorker control.
		/// </summary>
		/// <param name="sender">The source of the event.</param>
		/// <param name="e">The <see cref="T:System.ComponentModel.DoWorkEventArgs"/> instance containing the event data.</param>
		private void simulationWorker_DoWork(object sender, DoWorkEventArgs e)
		{

			Thread.CurrentThread.Priority = this.SimulationSettings.SimulatorThreadPriority;

			// Deny via CAS malicious process fragments
			processFragmentPermissionSet.Deny();


			while (!simulationWorker.CancellationPending)
			{
				lock (syncRoot)
				{

					// remove computers if their MTBF has passed
					meanTimeBetweenFailures++;
					if (meanTimeBetweenFailures >= simulationSettings.PaintableComputerFailureTime)
					{
						//reset mtbf time
						meanTimeBetweenFailures = 0;

						// remove random computer
						int random = RandomNumberGenerator.GetRandomNumberRange(0, this.computerList.Count);
						List<int> temp = new List<int>();
						temp.Add(random);
						DisableComputers(temp, true);
					}

					// perform simulationSettings.SimulatorCycle simulator cycles
					for (int i = 0; i < simulationSettings.SimulatorCycle; i++)
					{
						DoOneTick();
					}

					// feed values to sensors
					TraverseEnvironmentDataFeeders();

				}
			}
		}

		/// <summary>
		/// Handles the DoWork event of the simulationWorker control.
		/// </summary>
		/// <param name="sender">The source of the event.</param>
		/// <param name="e">The <see cref="T:System.ComponentModel.DoWorkEventArgs"/> instance containing the event data.</param>
		private void simulationWorkerWithEqualization_DoWork(object sender, DoWorkEventArgs e)
		{

			Thread.CurrentThread.Priority = this.SimulationSettings.SimulatorThreadPriority;

			Stopwatch sw = new Stopwatch();
			int sleepTime = 0;

			// Deny via CAS malicious process fragments
			processFragmentPermissionSet.Deny();

			while (!simulationWorker.CancellationPending)
			{
				// measure the time it takes to complete a cycle
				sw.Start();
				lock (syncRoot)
				{
					for (int i = 0; i < simulationSettings.SimulatorCycle; i++)
					{
						DoOneTick();
					}
				}
				sw.Stop();

				TraverseEnvironmentDataFeeders();

				// sleep for some time to prevent the initial perfromance boost
				sleepTime = (int)(equalizationCoeficient / (sw.ElapsedTicks + 1));
				sw.Reset();

				if (sleepTime > 10)
					Thread.Sleep(sleepTime);



			}
		}

		#endregion

		#region SimulatorBase Members

		/// <summary>
		/// Registers the environment data feed.
		/// </summary>
		/// <param name="environmentDataFeed">The environment data feed.</param>
		public void RegisterEnvironmentDataFeed(Type environmentDataFeeder)
		{

			if (environmentDataFeeder == null)
				throw new ArgumentNullException("environmentDataFeeder");

			if (!environmentDataFeeder.IsSubclassOf(typeof(EnvironmentDataFeeder)))
				throw new ArgumentException("This type does not inherit from EnvironmentDataFeeder!");

			// create the feeder and add it to the list
			object feeder = Activator.CreateInstance(environmentDataFeeder, new object[] { info.Environment.wallSize });
			List<int> sensorList = new List<int>();

			//traverse registered sensors to look for matching feeder/sensor and
			// add the matching sensor to the feeder's list of sensors
			for (int i = 0; i < computerList.Count; i++)
				if (computerList[i] is ISensor)
				{
					Type[] usingFeeders = ((ISensor)computerList[i]).DataFeeders;
					for (int j = 0; j < usingFeeders.Length; j++)
						if (usingFeeders[j] == environmentDataFeeder)
							sensorList.Add(i);
				}

			// add it to the list
			environment.Add((EnvironmentDataFeeder)feeder, sensorList);

			// log this call
			if (EngineLog.IsInfoEnabled)
				EngineLog.InfoFormat(UserMessages.EnvironmentFeederRegistered, ((FriendlyType)environmentDataFeeder).UserDefinedName);
		}

		/// <summary>
		/// Recreates the environment.
		/// </summary>
		private void RecreateEnvironment()
		{
			for (int i = 0; i < environment.Count; i++)
			{
				environment.DataList[i].Clear();

				//traverse computers to look for matching feeder/sensor and
				// add the matching sensor to the feeder's list of sensors
				for (int j = 0; j < computerList.Count; j++)
					if (computerList[j] is ISensor)
					{
						Type[] usingFeeders = ((ISensor)computerList[j]).DataFeeders;
						for (int k = 0; k < usingFeeders.Length; k++)
							if (usingFeeders[k] == environment.KeyList[i].GetType())
								environment.DataList[i].Add(j);
					}

			}

		}

		/// <summary>
		/// Adds additional computers.
		/// </summary>
		/// <param name="computerType">Type of the computer.</param>
		/// <param name="locations">The locations.</param>
		public void AddComputers(Type computerType, Point[] locations)
		{
			for (int i = 0; i < locations.Length; i++)
				AddComputer(computerType, locations[i]);
		}

		/// <summary>
		/// Adds an additional computer.
		/// </summary>
		/// <param name="sensorType">Type of the sensor.</param>
		/// <param name="location">The location.</param>
		public void AddComputer(Type computerType, Point location)
		{
			if (computerType == null)
				throw new ArgumentNullException("computerType");

			if (!computerType.IsSubclassOf(typeof(PaintableComputerBase)))
				throw new ArgumentException("This type is not a Paintable Computer!");

			// check for valid constructor
			ConstructorInfo constructor = computerType.GetConstructor(new Type[] { typeof(uint) });
			if (constructor == null)
				throw new ArgumentException("The computer doesn't have the right constructor: Constructor(Uint)");

			// create the computer
			object computer;
			try
			{
				computer = constructor.Invoke(new Object[] { simulationSettings.MaximumNeighbourCount });
			}
			catch
			{
				throw new Exception("Constructor failed!");
			}

			// create a computer info object
			SimpleComputerInfo newComputer = new SimpleComputerInfo();
			PaintableComputerBase paintableComputer = (PaintableComputerBase)computer;

			newComputer.Location = location;
			newComputer.ID = paintableComputer.ComputerId;
			newComputer.CompleteNeighboursIndices = new List<int>();
			newComputer.NeighboursIndices = new List<int>();

			// Schedule it
			newComputer.Tick = Convert.ToUInt32(simulationSettings.SimulatorCycle +
											   RandomNumberGenerator.GetRandomNumberRange(-simulationSettings.SimulatorCycleDeviation,
																						  simulationSettings.SimulatorCycleDeviation));

			// the index at wich the new computer will be inserted
			int insertionIndex = computerInfoList.Count;

			// add the computer info to the list
			computerInfoList.Add(newComputer);
			computerList.Add(paintableComputer);
			info.computerCount++;

			// Neighbour logic
			lock (syncRoot)
			{
				// search for neighbours
				for (int i = 0; i < computerInfoList.Count - 1; i++)
					if (PointIsInRadius(newComputer.Location, computerInfoList[i].Location, simulationSettings.WirelessRadius))
						newComputer.CompleteNeighboursIndices.Add(i);

				// add myself to the neighbours neighbour list
				for (int i = 0; i < newComputer.CompleteNeighboursIndices.Count; i++)
				{
					// my neighbour
					SimpleComputerInfo pc = computerInfoList[newComputer.CompleteNeighboursIndices[i]];

					pc.CompleteNeighboursIndices.Add(insertionIndex);
					int count = pc.CompleteNeighboursIndices.Count;

					// sort the newlly added Sensor in the other computers neighbourhood by distance
					for (int k = 0; k < count; k++)
						for (int l = k + 1; l < count; l++)
						{
							Point location1 = computerInfoList[pc.CompleteNeighboursIndices[l]].Location;
							Point location2 = computerInfoList[pc.CompleteNeighboursIndices[k]].Location;

							if (ComputerDistance(pc.Location, location1) < ComputerDistance(pc.Location, location2))
							{
								int index = pc.CompleteNeighboursIndices[l];
								pc.CompleteNeighboursIndices[l] = pc.CompleteNeighboursIndices[k];
								pc.CompleteNeighboursIndices[k] = index;
							}
						}

					// copy exactly simulationSettings.MaximumNeighbourCount PaC to the Neighbours list
					pc.NeighboursIndices.Clear();
					uint minIndex = (uint)pc.CompleteNeighboursIndices.Count;
					if (minIndex > simulationSettings.MaximumNeighbourCount)
						minIndex = simulationSettings.MaximumNeighbourCount;

					for (int r = 0; r < minIndex - 1; r++)
						pc.NeighboursIndices.Add(pc.CompleteNeighboursIndices[r]);

				}


				// copy exactly simulationSettings.MaximumNeighbourCount computers to the Neighbours list of the new computer
				newComputer.NeighboursIndices.Clear();

				uint minIndex2 = simulationSettings.MaximumNeighbourCount;
				if (newComputer.CompleteNeighboursIndices.Count <= simulationSettings.MaximumNeighbourCount)
					minIndex2 = (uint)newComputer.CompleteNeighboursIndices.Count;

				for (int r = 0; r < minIndex2; r++)
					newComputer.NeighboursIndices.Add(newComputer.CompleteNeighboursIndices[r]);



				// log this call
				if (EngineLog.IsInfoEnabled)
					EngineLog.InfoFormat(UserMessages.AdditionalComputerAdded, paintableComputer.ComputerId);

				// check if the computer has a sensor
				if (computerType.GetInterface(typeof(ISensor).Name) != null)
					RegisterSensor(paintableComputer, insertionIndex);




			}
		}

		/// <summary>
		/// Registers the sensor.
		/// </summary>
		/// <param name="computerType">Type of the computer.</param>
		/// <param name="paintableComputer">The paintable computer.</param>
		/// <param name="insertionIndex">Index of the insertion.</param>
		private void RegisterSensor(PaintableComputerBase paintableComputer, int sensorIndex)
		{
			// get the list of feeders this sensor reads from
			Type[] feeders = ((ISensor)paintableComputer).DataFeeders;
			for (int i = 0; i < environment.Count; i++)
			{
				Type feederType = environment.KeyList[i].GetType();

				for (int j = 0; j < feeders.Length; j++)
					if (feederType == feeders[j])
						environment.DataList[i].Add(sensorIndex);
			}

			// log this call
			if (EngineLog.IsInfoEnabled)
				EngineLog.Info(UserMessages.SensorRegistered);

		}

		/// <summary>
		/// Runs this instance.
		/// </summary>
		public override void Run()
		{
			if (!initialized)
				return;

			while (simulationWorker.CancellationPending) ;

			if (!simulationWorker.IsBusy)
			{
				simulationWorker.RunWorkerAsync();
				info.simulatorState = SimulatorStates.Running;
			}
		}

		/// <summary>
		/// Stops this instance.
		/// </summary>
		public override void Stop()
		{
			if (!initialized)
				return;

			if (simulationWorker.IsBusy)
			{
				simulationWorker.CancelAsync();

				info.simulatorState = SimulatorStates.Stopped;
				if (log.IsInfoEnabled) log.Info(UserMessages.SimulatorPaused);
			}
		}

		/// <summary>
		/// Steps this instance.
		/// </summary>
		public override void Step()
		{
			if (!initialized)
				return;

			//
			info.simulatorState = SimulatorStates.Running;

			for (int i = 0; i < simulationSettings.SimulatorCycle; i++)
			{
				DoOneTick();
			}

			//
			info.simulatorState = SimulatorStates.Stopped;
			if (log.IsInfoEnabled) log.Info(UserMessages.SimlatorStepComplete);
		}


		/// <summary>
		/// Initializes the simulator.
		/// </summary>
		public override void InitializeSimulator()
		{
			InitializeSimulator(typeof(GenericPaintableComputer));
		}

		/// <summary>
		/// Initializes the simulator with the specified paintable computer type.
		/// </summary>
		/// <param name="paintableComputerType">Type of the paintable computer.</param>
		public override void InitializeSimulator(Type paintableComputerType)
		{
			if (!paintableComputerType.IsSubclassOf(typeof(PaintableComputerBase)))
				throw new ArgumentException("paintableComputerType");

			if (log.IsInfoEnabled)
				log.Info(UserMessages.SimulatorInitializationDefaults);

			// Defaultne vrijenosti
			info.tickCount = 0;
			info.computerCount = 0;
			info.Communication.lostPacketCount = 0;
			info.Communication.pendingPacketCount = 0;
			info.Communication.totalPacketCount = 0;
			info.Environment.wallSize = new Size(0, 0);

			// create a new environment
			environment = new IndexedDictionary<EnvironmentDataFeeder, List<int>>();

			if (simulationSettings.LoadEqualization)
			{
				Stopwatch sw = new Stopwatch();

				double d = 0.0;
				long l = 0;
				int i = 0;

				sw.Start();
				// measue CPU power (approximate)
				for (int ind = 0; ind < 10000; ind++)
				{
					i += 100;
					d += 100.0;
					l += 100;

					i /= 3;
					d /= 3;
					l /= 3;

					i *= 7;
					d *= 7;
					l *= 7;

					i -= 11;
					d -= 11;
					l -= 11;
				}
				sw.Stop();

				// generate a coeficient used to balance the load
				equalizationCoeficient = sw.ElapsedTicks * 500;

				// attach the equalization simulatorWorker function
				this.simulationWorker.DoWork += new DoWorkEventHandler(this.simulationWorkerWithEqualization_DoWork);
			}
			else
			{
				// attach the normal simulatorWorker function
				this.simulationWorker.DoWork += new DoWorkEventHandler(this.simulationWorker_DoWork);
			}

			// Stvaranje PCjeva na zidu
			if (log.IsInfoEnabled)
				log.Info(String.Format(UserMessages.SimulatorInitialization, simulationSettings.MaximumComputerCount));

			SplashOnTheWall(simulationSettings.MaximumComputerCount, paintableComputerType);

			if (log.IsInfoEnabled)
				log.Info(String.Format(UserMessages.SimulatorInialized, info.computerCount));


			// Traenje susjeda
			uint neighbourCount = CalculateNeighbours(info.Environment.wallSize);
			if (log.IsInfoEnabled)
				log.Info(String.Format(UserMessages.SimulatorNeighbourCount, neighbourCount));

			// Sve je OK
			if (log.IsInfoEnabled)
				log.Info(UserMessages.SimulatorReady);

			//
			info.simulatorState = SimulatorStates.ReadyToRun;
			initialized = true;

		}

		/// <summary>
		/// </summary>
		public override void Terminate()
		{
			if (!initialized)
				return;

			if (log.IsInfoEnabled)
				log.Info(UserMessages.SimulatorThreadEnding);

			//Koristimo info varijablu za thread-safe rad
			simulationWorker.CancelAsync();

			if (log.IsInfoEnabled)
				log.Info(UserMessages.SimulatorThreadEnded);
		}

		/// <summary>
		/// Returns the simulator status
		/// </summary>
		/// <returns></returns>
		public override SimulatorInfo SimulatorStatus
		{
			get { return info; }
		}


		/// <summary>
		/// Removes the computers.
		/// </summary>
		/// <param name="computers">The computers.</param>
		/// <param name="remove">Ignored. Always removes them.</param>
		public override void DisableComputers(IList<int> computers, bool remove)
		{
			if (!initialized)
				return;

			List<int> temp = new List<int>(computers);

			temp.Sort();

			lock (syncRoot)
			{

				if (log.IsInfoEnabled)
					log.Info(UserMessages.SimulatorPaintableComputerMassDestruction);

				// store all pending packets
				List<RadioPacket> packets = new List<RadioPacket>(pendingRadioPackets.ToArray());

				// process all computers to be removed
				for (int i = temp.Count - 1; i >= 0; i--)
				{
					int removeAt = temp[i];

					if ((removeAt >= 0) && (removeAt < info.computerCount))
					{

						computerInfoList.RemoveAt(removeAt);
						computerList.RemoveAt(removeAt);

						for (int j = packets.Count - 1; j >= 0; j--)
						{
							if (packets[j].OriginComputerIndex == temp[i])
								packets.RemoveAt(j);

							// decrease indexes of origin indexes that are larger than the one we are removing
							if (packets[j].OriginComputerIndex > removeAt)
								packets[j].OriginComputerIndex -= 1;
						}

					}
				}

				// recalculate neighbours
				CalculateNeighbours(info.Environment.wallSize);

				// recreate sensors
				RecreateEnvironment();

				// update computer count
				info.computerCount = Convert.ToUInt32(computerInfoList.Count);


				// check for stray packets and update indexes above the index to be removed
				pendingRadioPackets = new Queue<RadioPacket>(packets);

			}
		}

		/// <summary>
		/// Invades the computer.
		/// </summary>
		/// <param name="computerIndex">Index of the computer.</param>
		/// <param name="processFragmentType">Type of the process fragment.</param>
		/// <param name="state">The state.</param>
		public override void InvadeComputer(int computerIndex, ProcessFragmentBase processFragment)
		{
			if (!initialized)
				return;

			// if none was chosen, insert into first computer
			if ((computerInfoList.Count <= computerIndex) && (computerIndex < 0))
				computerIndex = 0;

			if (RemotingServices.IsTransparentProxy(processFragment))
				throw new Exception("Passing a proxy!");

			// construct a new packet containing the created ProcessFragment copy
			ProcessFragmentPacket packet =
				new ProcessFragmentPacket(processFragment, 0, computerList[computerIndex].ComputerId);

			computerList[computerIndex].TransmissionAntenna = packet;
			computerInfoList[computerIndex].IsPortal = true;
		}

		/// <summary>
		/// Gets the computer list.
		/// </summary>
		/// <value>The computer list.</value>
		public override List<SimpleComputerInfo> ComputerList
		{
			get { return computerInfoList; }
		}

		/// <summary>
		/// Provides access to the simulation parameters
		/// </summary>
		/// <value>The simulation settings.</value>
		public override EngineSettings SimulationSettings
		{
			get { return simulationSettings; }
			set { simulationSettings = value; }
		}

		#endregion

		#region Homepage Colorizer

		/// <summary>
		/// Computers the colorizer.
		/// </summary>
		/// <param name="computer">The computer.</param>
		/// <returns></returns>
		private Color ComputerToColorConverter(PaintableComputerBase computer)
		{
			// THIS prevents process fragment colorizers from colorizing
			return homepageColorizerDelegate(computer.LocalHomepage);
		}

		/// <summary>
		/// Colorizes the paintable computer homepages.
		/// </summary>
		/// <param name="converter">The converter.</param>
		/// <returns>A list of colors matching the list of paintable computers</returns>
		public override List<Color> ColorizeHomepages(FriendlyType colorizer)
		{
			List<Color> retval = null;

			if (colorizer == null)
				homepageColorizer = defaultColorizer;
			else
			{
				try
				{
					homepageColorizer = (HomepageColorizer)Activator.CreateInstance(colorizer);
				}
				catch
				{
				}
			}

			// if none specified, use default
			if (homepageColorizer == null)
				homepageColorizer = defaultColorizer;

			homepageColorizerDelegate = new Converter<Homepage, Color>(homepageColorizer.Colorize);


			// perform a conversion of computers to color, which in turn calls the pfrag colorizter

			lock (syncRoot)
			{
				try
				{
					retval = computerList.ConvertAll<Color>(new Converter<PaintableComputerBase, Color>(ComputerToColorConverter));
				}
				catch
				{
					homepageColorizerDelegate = new Converter<Homepage, Color>(defaultColorizer.Colorize);
					homepageColorizer = defaultColorizer;
					retval = computerList.ConvertAll<Color>(new Converter<PaintableComputerBase, Color>(ComputerToColorConverter));
				}
			}

			return retval;
		}

		#endregion


		/// <summary>
		/// Traverses the environment data feeders and sets environment data to sensors.
		/// </summary>
		private void TraverseEnvironmentDataFeeders()
		{
			// feed data to sensors
			for (int i = 0; i < environment.Count; i++)
			{
				EnvironmentDataFeeder feeder = environment.KeyList[i];
				List<int> sensors = environment.DataList[i];

				for (int j = 0; j < sensors.Count; j++)
				{
					Point location = computerInfoList[sensors[j]].Location;

					try
					{
						// get value from feeder
						double value = feeder.GetValue(location, info.tickCount);

						// set the sensor 
						((ISensor)computerList[sensors[j]]).SetEnvironmentValue(feeder.GetType(), value);

					}
					catch (SecurityException ex)
					{
						// the feeder tried some action that is not allowed by the CAS set in the main simulator thread
						if (EngineLog.IsErrorEnabled)
							EngineLog.ErrorFormat("Datafeeder error ({0}): {1}", feeder.GetType(), ex.ToString());
					}
					catch
					{
						// other error
					}
				}
			}
		}

		/// <summary>
		/// Places Computers on wall
		/// </summary>
		/// <param name="maxComputerCount">Maximum computers on the wall</param>
		/// <param name="paintableComputerType">Type of the paintable computer.</param>
		private void SplashOnTheWall(uint maxComputerCount, Type paintableComputerType)
		{
			if (paintableComputerType == null)
				throw new ArgumentNullException("paintableComputerType");

			if (!paintableComputerType.IsSubclassOf(typeof(PaintableComputerBase)))
				throw new ArgumentException("The type of the paintableComputerType should inherit from PaintableComputerBase");

			int xi = Convert.ToInt32(Math.Sqrt(maxComputerCount) * simulationSettings.WallSizeRatio.Width / simulationSettings.WallSizeRatio.Height);
			int yi = Convert.ToInt32(maxComputerCount / xi);

			info.computerCount = Convert.ToUInt32(xi * yi);

			// Varijacije na temu
			int Vartiations = Convert.ToInt32(simulationSettings.SpaceBetweenPaintableComputers / 2.2);

			info.Environment.wallSize.Width = 0;
			info.Environment.wallSize.Height = 0;

			// Stvorimo listu PCova...
			computerInfoList = new List<SimpleComputerInfo>();
			computerList = new List<PaintableComputerBase>();

			// ... i popunimo ih
			for (int i = 1; i <= xi; i++)
				for (int j = 1; j <= yi; j++)
				{
					SimpleComputerInfo pc = new SimpleComputerInfo();
					PaintableComputerBase computer = null;

					// Male varijacije u slaganju
					pc.Location.X = Convert.ToInt32((i * simulationSettings.SpaceBetweenPaintableComputers) +
													RandomNumberGenerator.GetRandomNumberRange(-Vartiations, +Vartiations));
					pc.Location.Y = Convert.ToInt32((j * simulationSettings.SpaceBetweenPaintableComputers) +
													RandomNumberGenerator.GetRandomNumberRange(-Vartiations, +Vartiations));

					computer = (PaintableComputerBase)Activator.CreateInstance(paintableComputerType, new object[] { simulationSettings.MaximumNeighbourCount });

					// Schedule it
					pc.Tick = Convert.ToUInt32(simulationSettings.SimulatorCycle +
											   RandomNumberGenerator.GetRandomNumberRange(-simulationSettings.SimulatorCycleDeviation,
																						  simulationSettings.SimulatorCycleDeviation));

					// create neighbour lists
					pc.NeighboursIndices = new List<int>();
					pc.CompleteNeighboursIndices = new List<int>();

					int insertionIndex = computerInfoList.Count;

					// Add the computer info and the computer into the list
					computerInfoList.Add(pc);
					computerList.Add(computer);

					// check if the computer is a sensor
					if (computer is ISensor)
						RegisterSensor(computer, insertionIndex);


					// Osvjei veliinu zida
					if (info.Environment.wallSize.Height < pc.Location.Y)
						info.Environment.wallSize.Height = pc.Location.Y;
					if (info.Environment.wallSize.Width < pc.Location.X)
						info.Environment.wallSize.Width = pc.Location.X;


				}

			// Za sigurnost
			info.Environment.wallSize.Height += Convert.ToInt32(2 * simulationSettings.SpaceBetweenPaintableComputers);
			info.Environment.wallSize.Width += Convert.ToInt32(2 * simulationSettings.SpaceBetweenPaintableComputers);
		}


		/// <summary>
		/// Increments TickCount and
		/// </summary>
		/// <remarks>
		/// - Deliveres radio packets
		/// - Runs scheduled Cycle methods of Paintable Computers
		/// - ...
		/// </remarks>
		public void DoOneTick()
		{
			info.tickCount++;

			// Na poetku dostavimo pakete
			if (info.Communication.pendingPacketCount > 0)
			{
				RadioPacket rp = pendingRadioPackets.Peek();
				// ovo < je za svaki sluaj!
				while (rp.SendAfterTick <= info.tickCount)
				{
					SimpleComputerInfo originComputer = computerInfoList[rp.OriginComputerIndex];

					pendingRadioPackets.Dequeue();
					for (int i = 0; i < originComputer.NeighboursIndices.Count; i++)
					{
						if (RadioPacket.WillBeSent(simulationSettings.RadioLinkFailurePercent))
						{
							PaintableComputerBase pc = computerList[originComputer.NeighboursIndices[i]];
							pc.TransmissionAntenna = (NetworkPacket)rp.Payload;
							info.Communication.totalPacketCount++;
						}
						else
							info.Communication.lostPacketCount++;
					}

					// Ako se lista ispraznila, izai vani
					if (pendingRadioPackets.Count == 0)
						break;

					rp = pendingRadioPackets.Peek();
				}
				info.Communication.pendingPacketCount = Convert.ToUInt32(pendingRadioPackets.Count);
			}


			for (int i = 0; i < computerInfoList.Count; i++)
			{
				SimpleComputerInfo pcInfo = computerInfoList[i];
				PaintableComputerBase pc = computerList[i];

				// Is the PC scheduled for processing
				if (pcInfo.Tick <= info.tickCount)
				{
					// Cycle the computer
					pc.Cycle();

					// Ako ima sto za poslati
					while (true)
					{
						NetworkPacket packet = pc.TransmissionAntenna;
						if (packet != null)
						{
							RadioPacket rp = new RadioPacket();
							rp.OriginComputerIndex = i;
							rp.SendAfterTick = info.tickCount + simulationSettings.RadioLinkDelay;
							rp.Payload = packet;
							pendingRadioPackets.Enqueue(rp);
						}
						else
							break;
					}

					// Odredimo kad ce se ponovo dogoditi odrada
					pcInfo.Tick += Convert.ToUInt32(simulationSettings.SimulatorCycle +
												RandomNumberGenerator.GetRandomNumberRange(
													-simulationSettings.SimulatorCycleDeviation,
													simulationSettings.SimulatorCycleDeviation));
				}
				// Nije, idemo dalje...
			}
			info.Communication.pendingPacketCount = Convert.ToUInt32(pendingRadioPackets.Count);
		}


		/// <summary>
		/// Is point in radius of another point
		/// </summary>
		/// <param name="Center"></param>
		/// <param name="TestPoint"></param>
		/// <param name="Radius"></param>
		/// <returns></returns>
		private bool PointIsInRadius(Point Center, Point TestPoint, uint Radius)
		{
			int dx = Center.X - TestPoint.X;
			int dy = Center.Y - TestPoint.Y;
			int dr2 = dx * dx + dy * dy;

			return dr2 <= Radius * Radius;
		}

		/// <summary>
		/// Calculates distance between two points
		/// </summary>
		/// <param name="p1">The p1.</param>
		/// <param name="p2">The p2.</param>
		/// <returns></returns>
		private static int ComputerDistance(Point p1, Point p2)
		{
			int dx = p1.X - p2.X;
			int dy = p1.Y - p2.Y;
			int dr2 = dx * dx + dy * dy;

			return dr2;
		}

		#region Optimized Neighbour Search

		/// <summary>
		/// Does the circle.
		/// </summary>
		/// <param name="pc">The pc.</param>
		/// <param name="wallIndexMatrice">The wall index matrice.</param>
		/// <param name="LeftRight">The left right.</param>
		/// <param name="WallBoundaries">The wall boundaries.</param>
		/// <param name="size">The size.</param>
		/// <returns></returns>
		private uint DoCircle(SimpleComputerInfo pc, int[,] wallIndexMatrice, Point LeftRight, Size WallBoundaries, int size)
		{
			uint Return = Convert.ToUInt32(pc.NeighboursIndices.Count);

			size--;
			if ((LeftRight.X >= 0) && (LeftRight.X < WallBoundaries.Height))
				for (int i = LeftRight.Y; i <= (LeftRight.Y + size); i++)
					if ((i >= 0) && (i < WallBoundaries.Width))
					{
						if (wallIndexMatrice[LeftRight.X, i] == -1) continue;
						if (PointIsInRadius(pc.Location, computerInfoList[wallIndexMatrice[LeftRight.X, i]].Location,
											simulationSettings.WirelessRadius))
							pc.NeighboursIndices.Add(wallIndexMatrice[LeftRight.X, i]);
					}

			LeftRight.X += size;
			if ((LeftRight.X >= 0) && (LeftRight.X < WallBoundaries.Height))
				for (int i = LeftRight.Y; i <= (LeftRight.Y + size); i++)
					if ((i >= 0) && (i < WallBoundaries.Width))
					{
						if (wallIndexMatrice[LeftRight.X, i] == -1) continue;
						if (PointIsInRadius(pc.Location, computerInfoList[wallIndexMatrice[LeftRight.X, i]].Location,
											simulationSettings.WirelessRadius))
							pc.NeighboursIndices.Add(wallIndexMatrice[LeftRight.X, i]);
					}

			LeftRight.X -= size;
			LeftRight.X++;
			size -= 2;
			if ((LeftRight.Y >= 0) && (LeftRight.Y < WallBoundaries.Width))
				for (int i = LeftRight.X; i <= (LeftRight.X + size); i++)
					if ((i >= 0) && (i < WallBoundaries.Height))
					{
						if (wallIndexMatrice[i, LeftRight.Y] == -1) continue;
						if (PointIsInRadius(pc.Location, computerInfoList[wallIndexMatrice[i, LeftRight.Y]].Location,
											simulationSettings.WirelessRadius))
							pc.NeighboursIndices.Add(wallIndexMatrice[i, LeftRight.Y]);
					}


			LeftRight.Y += size + 2;
			if ((LeftRight.Y >= 0) && (LeftRight.Y < WallBoundaries.Width))
				for (int i = LeftRight.X; i <= (LeftRight.X + size); i++)
					if ((i >= 0) && (i < WallBoundaries.Height))
					{
						if (wallIndexMatrice[i, LeftRight.Y] == -1) continue;
						if (PointIsInRadius(pc.Location, computerInfoList[wallIndexMatrice[i, LeftRight.Y]].Location,
											simulationSettings.WirelessRadius))
							pc.NeighboursIndices.Add(wallIndexMatrice[i, LeftRight.Y]);
					}

			return Convert.ToUInt32(pc.NeighboursIndices.Count) - Return;
		}

		/// <summary>
		/// 
		/// </summary>
		/// <param name="WidthAndHeight"></param>
		/// <returns></returns>
		public uint CalculateNeighbours(Size wallSize)
		{
			// Matrix that holds the computer info indexes according to the computeInfoList list
			int[,] wallMatrix = new int[wallSize.Height, wallSize.Width];
			for (int i = 0; i < wallSize.Height; i++)
				for (int j = 0; j < wallSize.Width; j++)
					wallMatrix[i, j] = -1;

			// Set the index in the matrice according to the computer location
			for (int i = 0; i < computerInfoList.Count; i++)
				wallMatrix[computerInfoList[i].Location.Y, computerInfoList[i].Location.X] = i;


			// Sad su u reetci
			// Krui oko njih dok god se 2 kruga ne pojavi niti jedan novi susjed
			Point LeftRight = new Point();
			uint neighbours = 0;
			for (int i = 0; i < computerInfoList.Count; i++)
			{
				SimpleComputerInfo pc = computerInfoList[i];
				pc.NeighboursIndices.Clear();

				//Trenutni radius
				uint radius = 1;

				// Koliko ima PCona u redu
				int length = 3;

				// Gornji lijevi susjed
				Point StartPoint = new Point(pc.Location.Y, pc.Location.X);
				LeftRight.X = pc.Location.Y - 1;
				LeftRight.Y = pc.Location.X - 1;

				if (i == computerInfoList.Count - 1)
					radius = 1;

				// Kruimo poveavajui radijus do njegove granice
				uint NeighbourCount = 0;
				do
				{
					NeighbourCount += DoCircle(pc, wallMatrix, LeftRight, wallSize, length);

					// Novi gornji lijevim radius i duzina kvadrata
					radius++;
					length += 2;
					LeftRight.X--;
					LeftRight.Y--;

				} while ((radius <= SimulationSettings.WirelessRadius) && (NeighbourCount < SimulationSettings.MaximumNeighbourCount));

				// copy complete neighbour list
				pc.CompleteNeighboursIndices = new List<int>(pc.NeighboursIndices.ToArray());

				// Ako smo presli limit
				if (pc.NeighboursIndices.Count > SimulationSettings.MaximumNeighbourCount)
				{
					pc.NeighboursIndices.RemoveRange(Convert.ToInt32(SimulationSettings.MaximumNeighbourCount),
											  Convert.ToInt32(pc.NeighboursIndices.Count - SimulationSettings.MaximumNeighbourCount));
				}

				neighbours += Convert.ToUInt32(pc.NeighboursIndices.Count);

			}

			return neighbours;
		}


		#endregion Optimized Neighbour Search


		/// <summary>
		/// Gets the computer.
		/// </summary>
		/// <param name="index">The index.</param>
		/// <returns></returns>
		public PaintableComputerBase GetComputer(int index)
		{
			if ((index < 0) && (index >= computerList.Count))
				return null;

			return computerList[index];
		}
	}
}