/*
	Paintable Computing Simulation

	License

	This software (Paintable Computing Simulation and all parts of its 
	source code) is provided "as is", without any warranty and it is 
	the property of its respective authors. Use of this software is 
	free to everyone, except in commercial and/or military purposes 
	without written permission of its authors. Disassembling the 
	executable code of this software is forbidden for any purpose 
	without written permission of its authors.

	If you intend to use any part of the source code of this software 
	or to develop this software any further, you must include this 
	license in every source file and every About box, without changing 
	it; only appending the license is allowed. And if you use any part 
	of the source code of this software, or develop it further, please 
	send an e-mail to its authors.

	Authors of this software are not responsible for any consequences 
	which use of this software, use of any part of its source or its 
	further development may cause.

	All software and parts of source code used in the development of 
	this software are property of their respective owners.

	Good luck,
	Authors

	Authors (alphabetically):
	Fabijan Lukin
		fabijan@gmail.com
		fabijan.lukin@fer.hr
	Fran Pregernik
		fran.pregernik@gmail.com
		fran.pregernik@fer.hr
	Tomislav Sukser
		tomislav.sukser@gmail.com
		tomislav.sukser@zg.t-com.hr
		tomislav.sukser@fer.hr


	Note:
	Paintable computing was firstly introduced in Programming a 
	Paintable Computer written by William Joseph Butera.


    Copyright (c) 2005,2006. All rights reserved.
    4-6/2006

 */
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Drawing;
using System.Runtime.Remoting.Lifetime;
using System.Security;
using System.Security.Permissions;
using PaintableSimulator.Engine.PaintableComputers.OperatingSystem;
using PaintableSimulator.Engine.PaintableComputers.Sensors;
using System.Reflection;
using System.IO;
using System.IO.Compression;
using System.Diagnostics;

namespace PaintableSimulator.Engine
{

	#region Default Colorizer

	/// <summary>
	/// 
	/// </summary>
	public class DefaultColorizer : HomepageColorizer
	{

		/// <summary>
		/// Colorizes the specified homepage.
		/// </summary>
		/// <param name="homepage">The homepage.</param>
		/// <returns></returns>
		public override Color Colorize(Homepage homepage)
		{
			Color retval = Color.WhiteSmoke;
			return retval;
		}
	}

	#endregion

	[Serializable()]
	public class SimpleComputerInfo
	{
		public bool IsPortal;
		public Point Location;
		public uint ID;
		public uint Tick;

		public List<int> NeighboursIndices;
		internal List<int> CompleteNeighboursIndices;

	}
	
	public abstract class SimulatorBase : MarshalByRefObject
	{

		// Create a logger for use in this class
		protected static readonly log4net.ILog log = log4net.LogManager.GetLogger("Simulator");
		// NOTE that using System.Reflection.MethodBase.GetCurrentMethod().DeclaringType
		// is equivalent to typeof(LoggingExample) but is more portable
		// i.e. you can copy the code directly into another class without
		// needing to edit the code.

		protected PluginLoader pluginLoader;

		public override object InitializeLifetimeService()
		{
			ILease lease = (ILease) base.InitializeLifetimeService();
			if (lease.CurrentState == LeaseState.Initial)
			{
				lease.InitialLeaseTime = TimeSpan.FromMinutes(5);
				lease.SponsorshipTimeout = TimeSpan.FromMinutes(2);
				lease.RenewOnCallTime = TimeSpan.FromSeconds(10);
			}
			return lease;
		}


		/// <summary>
		/// 
		/// </summary>
		protected PermissionSet processFragmentPermissionSet = new PermissionSet(PermissionState.None);


		/// <summary>
		/// Initializes a new instance of the <see cref="T:SimulatorBase"/> class.
		/// </summary>
		protected SimulatorBase()
		{
			// load plugin loader
			pluginLoader = new PluginLoader();

			processFragmentPermissionSet.AddPermission(new FileIOPermission(PermissionState.Unrestricted));
			processFragmentPermissionSet.AddPermission(new EnvironmentPermission(PermissionState.Unrestricted));
			processFragmentPermissionSet.AddPermission(new FileDialogPermission(PermissionState.Unrestricted));
			processFragmentPermissionSet.AddPermission(new FileIOPermission(PermissionState.Unrestricted));
			processFragmentPermissionSet.AddPermission(new IsolatedStorageFilePermission(PermissionState.Unrestricted));
			processFragmentPermissionSet.AddPermission(new KeyContainerPermission(PermissionState.Unrestricted));
			processFragmentPermissionSet.AddPermission(new PublisherIdentityPermission(PermissionState.Unrestricted));
			processFragmentPermissionSet.AddPermission(new ReflectionPermission(PermissionState.Unrestricted));
			processFragmentPermissionSet.AddPermission(new RegistryPermission(PermissionState.Unrestricted));
			processFragmentPermissionSet.AddPermission(new StorePermission(PermissionState.Unrestricted));
			processFragmentPermissionSet.AddPermission(new UIPermission(PermissionState.Unrestricted));
			processFragmentPermissionSet.AddPermission(new UrlIdentityPermission(PermissionState.Unrestricted));
		}

		/// <summary>
		/// Gets the plugins.
		/// </summary>
		/// <value>The plugins.</value>
		public PluginLoader Plugins
		{
			get
			{
				return pluginLoader;
			}

		}

		/// <summary>
		/// Runs this instance.
		/// </summary>
		public abstract void Run();

		/// <summary>
		/// Stops this instance.
		/// </summary>
		public abstract void Stop();

		/// <summary>
		/// Steps this instance.
		/// </summary>
		public abstract void Step();


		/// <summary>
		/// Initializes the simulator.
		/// </summary>
		public abstract void InitializeSimulator();

		/// <summary>
		/// Initializes the simulator with the specified paintable computer type.
		/// </summary>
		/// <param name="paintableComputerType">Type of the paintable computer.</param>
		public abstract void InitializeSimulator(Type paintableComputerType);

		/// <summary>
		/// 
		/// </summary>
		public abstract void Terminate();

		/// <summary>
		/// Returns the simulator status
		/// </summary>
		/// <returns></returns>
		public abstract SimulatorInfo SimulatorStatus { get; }

		/// <summary>
		/// Disables the computers or removes it.
		/// </summary>
		/// <param name="computers">The computers.</param>
		/// <param name="remove">if set to <c>true</c> the compters will be removed.</param>
		public abstract void DisableComputers(IList<int> computers, bool remove);

		/// <summary>
		/// Invades the computer.
		/// </summary>
		/// <param name="computerIndex">Index of the computer.</param>
		/// <param name="processFragment">The process fragment.</param>
		public abstract void InvadeComputer(int computerIndex, ProcessFragmentBase processFragment);


		/// <summary>
		/// Gets the computer list.
		/// </summary>
		/// <value>The computer list.</value>
		public abstract List<SimpleComputerInfo> ComputerList { get; }


		/// <summary>
		/// Provides access to the simulation parameters
		/// </summary>
		/// <value>The simulation settings.</value>
		public abstract EngineSettings SimulationSettings { get; set; }


		/// <summary>
		/// Colorization delegate
		/// </summary>
		protected Converter<Homepage, Color> homepageColorizerDelegate = null;

		/// <summary>
		/// Colorizes the paintable computer homepages.
		/// </summary>
		/// <param name="colorizer">The colorizer.</param>
		/// <returns>
		/// A list of colors matching the list of paintable computers
		/// </returns>
		public abstract List<Color> ColorizeHomepages(FriendlyType colorizer);

		/// <summary>
		/// The default Homepage colorizer.
		/// </summary>
		protected HomepageColorizer defaultColorizer = new DefaultColorizer();

		/// <summary>
		/// The selected Homepage colorizer
		/// </summary>
		protected HomepageColorizer homepageColorizer = null;

	}
}